"""
Package for working with JSON-format configuration files.
"""
from ._Absent import Absent
from ._Configuration import Configuration
from ._OptionallyPresent import OptionallyPresent
from ._StrictConfiguration import StrictConfiguration
from ._SubConfiguration import SubConfiguration
