"""
Package specifying interfaces for objects/classes which can perform JSON validation.
"""
from ._JSONValidatorClass import JSONValidatorClass
from ._JSONValidatorInstance import JSONValidatorInstance
