"""
Package for meta-programming constructs.
"""
from ._functions import has_been_overridden, get_class_from_function, is_class_function, unbind, does_not_raise
from ._metadata import with_metadata, get_metadata, has_metadata
