"""
Package for interfaces supporting JSON serialisation.
"""
from ._error import JSONSerialiseError
from ._JSONBiserialisable import JSONBiserialisable
from ._JSONDecodable import JSONDecodable
from ._JSONDeserialisable import JSONDeserialisable
from ._JSONEncodable import JSONEncodable
from ._JSONSerialisable import JSONSerialisable
from ._JSONValidatedBiserialisable import JSONValidatedBiserialisable
